const EMPTY_STRING = '';
const UTF8 = 'utf8';

const isUndefined = (thing) => thing == void 0;
const ifNotUndefined = (value, then, otherwise) =>
  isUndefined(value) ? otherwise?.() : then(value);
const slice = (arrayOrString, start, end) => arrayOrString.slice(start, end);

const arrayReduce = (array, cb, initial) => array.reduce(cb, initial);

const collSizeN = (collSizer) => (coll) =>
  arrayReduce(collValues(coll), (total, coll2) => total + collSizer(coll2), 0);
const collSize = (coll) => coll?.size ?? 0;
const collSize2 = collSizeN(collSize);
const collHas = (coll, keyOrValue) => coll?.has(keyOrValue) ?? false;
const collIsEmpty = (coll) => isUndefined(coll) || collSize(coll) == 0;
const collValues = (coll) => [...(coll?.values() ?? [])];
const collClear = (coll) => coll.clear();
const collForEach = (coll, cb) => coll?.forEach(cb);
const collDel = (coll, keyOrValue) => coll?.delete(keyOrValue);

const mapNew = (entries) => new Map(entries);
const mapKeys = (map) => [...(map?.keys() ?? [])];
const mapGet = (map, key) => map?.get(key);
const mapForEach = (map, cb) =>
  collForEach(map, (value, key) => cb(key, value));
const mapSet = (map, key, value) =>
  isUndefined(value) ? (collDel(map, key), map) : map?.set(key, value);
const mapEnsure = (map, key, getDefaultValue, hadExistingValue) => {
  if (!collHas(map, key)) {
    mapSet(map, key, getDefaultValue());
  } else {
    hadExistingValue?.(mapGet(map, key));
  }
  return mapGet(map, key);
};

const MESSAGE_SEPARATOR = '\n';

const PATH_REGEX = /\/([^?]*)/;
const createWsServer = (webSocketServer) => {
  const clientsByPath = mapNew();
  webSocketServer.on('connection', (webSocket, request) =>
    ifNotUndefined(request.url?.match(PATH_REGEX), ([path]) =>
      ifNotUndefined(request.headers['sec-websocket-key'], (clientId) => {
        const clients = mapEnsure(clientsByPath, path, mapNew);
        mapSet(clients, clientId, webSocket);
        webSocket.on('message', (data) => {
          const payload = data.toString(UTF8);
          const splitAt = payload.indexOf(MESSAGE_SEPARATOR);
          if (splitAt !== -1) {
            const toClientId = slice(payload, 0, splitAt);
            const message = slice(payload, splitAt + 1);
            toClientId === EMPTY_STRING
              ? mapForEach(clients, (otherClientId, otherWebSocket) =>
                  otherClientId != clientId
                    ? otherWebSocket.send(
                        clientId + MESSAGE_SEPARATOR + message,
                      )
                    : 0,
                )
              : mapGet(clients, toClientId)?.send(
                  clientId + MESSAGE_SEPARATOR + message,
                );
          }
        });
        webSocket.on('close', () => {
          collDel(clients, clientId);
          if (collIsEmpty(clients)) {
            collDel(clientsByPath, path);
          }
        });
      }),
    ),
  );
  const getWebSocketServer = () => webSocketServer;
  const getPaths = () => mapKeys(clientsByPath);
  const getClientIds = (path) => mapKeys(mapGet(clientsByPath, path));
  const getStats = () => ({
    paths: collSize(clientsByPath),
    clients: collSize2(clientsByPath),
  });
  const destroy = () => {
    webSocketServer.close();
    collClear(clientsByPath);
  };
  return {
    getWebSocketServer,
    getPaths,
    getClientIds,
    getStats,
    destroy,
  };
};

export {createWsServer};
